//
// Cforall Version 1.0.0 Copyright (C) 2015 University of Waterloo
//
// The contents of this file are covered under the licence agreement in the
// file "LICENCE" distributed with Cforall.
//
// Lvalue.cc --
//
// Author           : Richard C. Bilson
// Created On       : Mon May 18 07:44:20 2015
// Last Modified By : Peter A. Buhr
// Last Modified On : Fri Mar 17 09:11:18 2017
// Update Count     : 5
//

#include <cassert>                       // for strict_dynamic_cast
#include <string>                        // for string

#include "Common/PassVisitor.h"
#include "GenPoly.h"                     // for isPolyType
#include "Lvalue.h"

#include "Parser/LinkageSpec.h"          // for Spec, isBuiltin, Intrinsic
#include "ResolvExpr/TypeEnvironment.h"  // for AssertionSet, OpenVarSet
#include "ResolvExpr/Unify.h"            // for unify
#include "ResolvExpr/typeops.h"
#include "SymTab/Autogen.h"
#include "SymTab/Indexer.h"              // for Indexer
#include "SynTree/Declaration.h"         // for Declaration, FunctionDecl
#include "SynTree/Expression.h"          // for Expression, ConditionalExpr
#include "SynTree/Mutator.h"             // for mutateAll, Mutator
#include "SynTree/Statement.h"           // for ReturnStmt, Statement (ptr o...
#include "SynTree/Type.h"                // for PointerType, Type, FunctionType
#include "SynTree/Visitor.h"             // for Visitor, acceptAll

#if 0
#define PRINT(x) x
#else
#define PRINT(x)
#endif

namespace GenPoly {
	namespace {
		// TODO: fold this into the general createDeref function??
		Expression * mkDeref( Expression * arg ) {
			if ( SymTab::dereferenceOperator ) {
				VariableExpr * deref = new VariableExpr( SymTab::dereferenceOperator );
				deref->result = new PointerType( Type::Qualifiers(), deref->result );
				Type * base = InitTweak::getPointerBase( arg->result );
				assertf( base, "expected pointer type in dereference (type was %s)", toString( arg->result ).c_str() );
				ApplicationExpr * ret = new ApplicationExpr( deref, { arg } );
				delete ret->result;
				ret->result = base->clone();
				ret->result->set_lvalue( true );
				return ret;
			} else {
				return UntypedExpr::createDeref( arg );
			}
		}

		struct ReferenceConversions final {
			Expression * postmutate( CastExpr * castExpr );
			Expression * postmutate( AddressExpr * addrExpr );
		};

		/// Intrinsic functions that take reference parameters don't REALLY take reference parameters -- their reference arguments must always be implicitly dereferenced.
		struct FixIntrinsicArgs final {
			Expression * postmutate( ApplicationExpr * appExpr );
		};

		struct FixIntrinsicResult final : public WithGuards {
			Expression * postmutate( ApplicationExpr * appExpr );
			void premutate( FunctionDecl * funcDecl );
			bool inIntrinsic = false;
		};

		/// Replace reference types with pointer types
		struct ReferenceTypeElimination final {
			Type * postmutate( ReferenceType * refType );
		};

		/// GCC-like Generalized Lvalues (which have since been removed from GCC)
		/// https://gcc.gnu.org/onlinedocs/gcc-3.4.6/gcc/Lvalues.html#Lvalues
		/// Replaces &(a,b) with (a, &b), &(a ? b : c) with (a ? &b : &c)
		struct GeneralizedLvalue final : public WithVisitorRef<GeneralizedLvalue> {
			Expression * postmutate( AddressExpr * addressExpr );
			Expression * postmutate( MemberExpr * memExpr );

			template<typename Func>
			Expression * applyTransformation( Expression * expr, Expression * arg, Func mkExpr );
		};

		/// Removes redundant &*/*& pattern that this pass can generate
		struct CollapseAddrDeref final {
			Expression * postmutate( AddressExpr * addressExpr );
			Expression * postmutate( ApplicationExpr * appExpr );
		};

		struct AddrRef final : public WithGuards {
			void premutate( AddressExpr * addrExpr );
			Expression * postmutate( AddressExpr * addrExpr );
			void premutate( Expression * expr );

			bool first = true;
			bool current = false;
			int refDepth = 0;
		};
	} // namespace

	static bool referencesEliminated = false;
	// used by UntypedExpr::createDeref to determine whether result type of dereference should be ReferenceType or value type.
	bool referencesPermissable() {
		return ! referencesEliminated;
	}

	void convertLvalue( std::list< Declaration* >& translationUnit ) {
		PassVisitor<ReferenceConversions> refCvt;
		PassVisitor<ReferenceTypeElimination> elim;
		PassVisitor<GeneralizedLvalue> genLval;
		PassVisitor<FixIntrinsicArgs> fixer;
		PassVisitor<CollapseAddrDeref> collapser;
		PassVisitor<AddrRef> addrRef;
		PassVisitor<FixIntrinsicResult> intrinsicResults;
		mutateAll( translationUnit, intrinsicResults );
		mutateAll( translationUnit, addrRef );
		mutateAll( translationUnit, refCvt );
		mutateAll( translationUnit, fixer );
		mutateAll( translationUnit, collapser );
		mutateAll( translationUnit, genLval );
		mutateAll( translationUnit, elim );  // last because other passes need reference types to work

		// from this point forward, no other pass should create reference types.
		referencesEliminated = true;
	}

	Expression * generalizedLvalue( Expression * expr ) {
		PassVisitor<GeneralizedLvalue> genLval;
		return expr->acceptMutator( genLval );
	}

	namespace {
		// true for intrinsic function calls that return a reference
		bool isIntrinsicReference( Expression * expr ) {
			if ( UntypedExpr * untyped = dynamic_cast< UntypedExpr * >( expr ) ) {
				std::string fname = InitTweak::getFunctionName( untyped );
				// known intrinsic-reference prelude functions
				return fname == "*?" || fname == "?[?]";
			} else if ( ApplicationExpr * appExpr = dynamic_cast< ApplicationExpr * > ( expr ) ) {
				if ( DeclarationWithType * func = InitTweak::getFunction( appExpr ) ) {
					// use type of return variable rather than expr result type, since it may have been changed to a pointer type
					FunctionType * ftype = GenPoly::getFunctionType( func->get_type() );
					Type * ret = ftype->get_returnVals().empty() ? nullptr : ftype->get_returnVals().front()->get_type();
					return func->get_linkage() == LinkageSpec::Intrinsic && dynamic_cast<ReferenceType *>( ret );
				}
			}
			return false;
		}

		Expression * FixIntrinsicResult::postmutate( ApplicationExpr * appExpr ) {
			if ( isIntrinsicReference( appExpr ) ) {
				// eliminate reference types from intrinsic applications - now they return lvalues
				Type * result = appExpr->get_result();
				appExpr->set_result( result->stripReferences()->clone() );
				appExpr->get_result()->set_lvalue( true );
				if ( ! inIntrinsic ) {
					// when not in an intrinsic function, add a cast to
					// don't add cast when in an intrinsic function, since they already have the cast
					Expression * ret = new CastExpr( appExpr, result );
					ret->set_env( appExpr->get_env() );
					appExpr->set_env( nullptr );
					return ret;
				}
				delete result;
			}
			return appExpr;
		}

		void FixIntrinsicResult::premutate( FunctionDecl * funcDecl ) {
			GuardValue( inIntrinsic );
			inIntrinsic =  funcDecl->linkage == LinkageSpec::Intrinsic;
		}

		Expression * FixIntrinsicArgs::postmutate( ApplicationExpr * appExpr ) {
			// intrinsic functions don't really take reference-typed parameters, so they require an implicit dereference on their arguments.
			if ( DeclarationWithType * function = InitTweak::getFunction( appExpr ) ) {
				FunctionType * ftype = GenPoly::getFunctionType( function->get_type() );
				assertf( ftype, "Function declaration does not have function type." );
				// can be of differing lengths only when function is variadic
				assertf( ftype->get_parameters().size() == appExpr->get_args().size() || ftype->get_isVarArgs(), "ApplicationExpr args do not match formal parameter type." );


				unsigned int i = 0;
				const unsigned int end = ftype->get_parameters().size();
				for ( auto p : unsafe_group_iterate( appExpr->get_args(), ftype->get_parameters() ) ) {
					if (i == end) break;
					Expression *& arg = std::get<0>( p );
					Type * formal = std::get<1>( p )->get_type();
					PRINT(
						std::cerr << "pair<0>: " << arg << std::endl;
						std::cerr << "pair<1>: " << formal << std::endl;
					)
					if ( dynamic_cast<ReferenceType*>( formal ) ) {
						if ( isIntrinsicReference( arg ) ) { // do not combine conditions, because that changes the meaning of the else if
							if ( function->get_linkage() != LinkageSpec::Intrinsic ) { // intrinsic functions that turn pointers into references
								// if argument is dereference or array subscript, the result isn't REALLY a reference, so it's not necessary to fix the argument
								PRINT(
									std::cerr << "===is intrinsic arg in non-intrinsic call - adding address" << std::endl;
								)
								arg = new AddressExpr( arg );
							}
						} else if ( function->get_linkage() == LinkageSpec::Intrinsic ) {
							// std::cerr << "===adding deref to arg" << std::endl;
							// if the parameter is a reference, add a dereference to the reference-typed argument.
							Type * baseType = InitTweak::getPointerBase( arg->get_result() );
							assertf( baseType, "parameter is reference, arg must be pointer or reference: %s", toString( arg->get_result() ).c_str() );
							PointerType * ptrType = new PointerType( Type::Qualifiers(), baseType->clone() );
							delete arg->get_result();
							arg->set_result( ptrType );
							arg = mkDeref( arg );
						}
					}
					++i;
				}
			}
			return appExpr;
		}

		// idea: &&&E: get outer &, inner &
		// at inner &, record depth D of reference type
		// at outer &, add D derefs.
		void AddrRef::premutate( Expression * ) {
			GuardValue( current );
			GuardValue( first );
			current = false;
			first = true;
		}

		void AddrRef::premutate( AddressExpr * ) {
			GuardValue( current );
			GuardValue( first );
			current = first;
			first = false;
			if ( current ) {
				GuardValue( refDepth );
				refDepth = 0;
			}
		}

		Expression * AddrRef::postmutate( AddressExpr * addrExpr ) {
			if ( refDepth == 0 ) {
				if ( ! isIntrinsicReference( addrExpr->get_arg() ) ) {
					// try to avoid ?[?]
					refDepth = addrExpr->get_arg()->get_result()->referenceDepth();
				}
			}
			if ( current ) {
				Expression * ret = addrExpr;
				while ( refDepth ) {
					ret = mkDeref( ret );
					refDepth--;
				}
				return ret;
			}
			return addrExpr;
		}

		Expression * ReferenceConversions::postmutate( AddressExpr * addrExpr ) {
			// Inner expression may have been lvalue to reference conversion, which becomes an address expression.
			// In this case, remove the outer address expression and return the argument.
			// TODO: It's possible that this might catch too much and require a more sophisticated check.
			return addrExpr;
		}

		Expression * ReferenceConversions::postmutate( CastExpr * castExpr ) {
			// xxx - is it possible to convert directly between reference types with a different base? E.g.,
			//   int x;
			//   (double&)x;
			// At the moment, I am working off of the assumption that this is illegal, thus the cast becomes redundant
			// after this pass, so trash the cast altogether. If that changes, care must be taken to insert the correct
			// pointer casts in the right places.

			// conversion to reference type
			if ( ReferenceType * refType = dynamic_cast< ReferenceType * >( castExpr->get_result() ) ) {
				(void)refType;
				if ( ReferenceType * otherRef = dynamic_cast< ReferenceType * >( castExpr->get_arg()->get_result() ) ) {
					// nothing to do if casting from reference to reference.
					(void)otherRef;
					PRINT( std::cerr << "convert reference to reference -- nop" << std::endl; )
					if ( isIntrinsicReference( castExpr->get_arg() ) ) {
						Expression * callExpr = castExpr->get_arg();
						PRINT(
							std::cerr << "but arg is deref -- &" << std::endl;
							std::cerr << callExpr << std::endl;
						)
						callExpr = new AddressExpr( callExpr ); // this doesn't work properly for multiple casts
						delete callExpr->get_result();
						callExpr->set_result( refType->clone() );
						// move environment out to new top-level
						callExpr->set_env( castExpr->get_env() );
						castExpr->set_arg( nullptr );
						castExpr->set_env( nullptr );
						delete castExpr;
						return callExpr;
					}
					int depth1 = refType->referenceDepth();
					int depth2 = otherRef->referenceDepth();
					int diff = depth1-depth2;
					if ( diff == 0 ) {
						// conversion between references of the same depth
						assertf( depth1 == depth2, "non-intrinsic reference with cast of reference to reference not yet supported: %d %d %s", depth1, depth2, toString( castExpr ).c_str() );
						PRINT( std::cerr << castExpr << std::endl; )
						return castExpr;
					} else if ( diff < 0 ) {
						// conversion from reference to reference with less depth (e.g. int && -> int &): add dereferences
						Expression * ret = castExpr->arg;
						for ( int i = 0; i < diff; ++i ) {
							ret = mkDeref( ret );
						}
						ret->env = castExpr->env;
						delete ret->result;
						ret->result = castExpr->result;
						ret->result->set_lvalue( true ); // ensure result is lvalue
						castExpr->env = nullptr;
						castExpr->arg = nullptr;
						castExpr->result = nullptr;
						delete castExpr;
						return ret;
					} else if ( diff > 0 ) {
						// conversion from reference to reference with more depth (e.g. int & -> int &&): add address-of
						Expression * ret = castExpr->arg;
						for ( int i = 0; i < diff; ++i ) {
							ret = new AddressExpr( ret );
						}
						ret->env = castExpr->env;
						delete ret->result;
						ret->result = castExpr->result;
						castExpr->env = nullptr;
						castExpr->arg = nullptr;
						castExpr->result = nullptr;
						delete castExpr;
						return ret;
					}

					assertf( depth1 == depth2, "non-intrinsic reference with cast of reference to reference not yet supported: %d %d %s", depth1, depth2, toString( castExpr ).c_str() );
					PRINT( std::cerr << castExpr << std::endl; )
					return castExpr;
				} else if ( castExpr->arg->result->get_lvalue() ) {
					// conversion from lvalue to reference
					// xxx - keep cast, but turn into pointer cast??
					// xxx - memory
					PRINT(
						std::cerr << "convert lvalue to reference -- &" << std::endl;
						std::cerr << castExpr->arg << std::endl;
					)
					AddressExpr * ret = new AddressExpr( castExpr->arg );
					if ( refType->base->get_qualifiers() != castExpr->arg->result->get_qualifiers() ) {
						// must keep cast if cast-to type is different from the actual type
						castExpr->arg = ret;
						return castExpr;
					}
					ret->env = castExpr->env;
					delete ret->result;
					ret->result = castExpr->result;
					castExpr->env = nullptr;
					castExpr->arg = nullptr;
					castExpr->result = nullptr;
					delete castExpr;
					return ret;
				} else {
					// rvalue to reference conversion -- introduce temporary
				}
				assertf( false, "Only conversions to reference from lvalue are currently supported: %s", toString( castExpr ).c_str() );
			} else if ( ReferenceType * refType = dynamic_cast< ReferenceType * >( castExpr->arg->result ) ) {
				(void)refType;
				// conversion from reference to rvalue
				PRINT(
					std::cerr << "convert reference to rvalue -- *" << std::endl;
					std::cerr << "was = " << castExpr << std::endl;
				)
				Expression * ret = castExpr->arg;
				TypeSubstitution * env = castExpr->env;
				castExpr->set_env( nullptr );
				if ( ! isIntrinsicReference( ret ) ) {
					// dereference if not already dereferenced
					ret = mkDeref( ret );
				}
				if ( ResolvExpr::typesCompatibleIgnoreQualifiers( castExpr->result, castExpr->arg->result->stripReferences(), SymTab::Indexer() ) ) {
					// can remove cast if types are compatible, changing expression type to value type
					ret->result = castExpr->result->clone();
					ret->result->set_lvalue( true );  // ensure result is lvalue
					castExpr->arg = nullptr;
					delete castExpr;
				} else {
					// must keep cast if types are different
					castExpr->arg = ret;
					ret = castExpr;
				}
				ret->set_env( env );
				PRINT( std::cerr << "now: " << ret << std::endl; )
				return ret;
			}
			return castExpr;
		}

		Type * ReferenceTypeElimination::postmutate( ReferenceType * refType ) {
			Type * base = refType->get_base();
			Type::Qualifiers qualifiers = refType->get_qualifiers();
			refType->set_base( nullptr );
			delete refType;
			return new PointerType( qualifiers, base );
		}

		template<typename Func>
		Expression * GeneralizedLvalue::applyTransformation( Expression * expr, Expression * arg, Func mkExpr ) {
			if ( CommaExpr * commaExpr = dynamic_cast< CommaExpr * >( arg ) ) {
				Expression * arg1 = commaExpr->get_arg1()->clone();
				Expression * arg2 = commaExpr->get_arg2()->clone();
				Expression * ret = new CommaExpr( arg1, mkExpr( arg2 )->acceptMutator( *visitor ) );
				ret->set_env( expr->get_env() );
				expr->set_env( nullptr );
				delete expr;
				return ret;
			} else if ( ConditionalExpr * condExpr = dynamic_cast< ConditionalExpr * >( arg ) ) {
				Expression * arg1 = condExpr->get_arg1()->clone();
				Expression * arg2 = condExpr->get_arg2()->clone();
				Expression * arg3 = condExpr->get_arg3()->clone();
				ConditionalExpr * ret = new ConditionalExpr( arg1, mkExpr( arg2 )->acceptMutator( *visitor ), mkExpr( arg3 )->acceptMutator( *visitor ) );
				ret->set_env( expr->get_env() );
				expr->set_env( nullptr );
				delete expr;

				// conditional expr type may not be either of the argument types, need to unify
				using namespace ResolvExpr;
				Type* commonType = nullptr;
				TypeEnvironment newEnv;
				AssertionSet needAssertions, haveAssertions;
				OpenVarSet openVars;
				unify( ret->get_arg2()->get_result(), ret->get_arg3()->get_result(), newEnv, needAssertions, haveAssertions, openVars, SymTab::Indexer(), commonType );
				ret->set_result( commonType ? commonType : ret->get_arg2()->get_result()->clone() );
				return ret;
			}
			return expr;
		}

		Expression * GeneralizedLvalue::postmutate( MemberExpr * memExpr ) {
			return applyTransformation( memExpr, memExpr->get_aggregate(), [=]( Expression * aggr ) { return new MemberExpr( memExpr->get_member(), aggr ); } );
		}

		Expression * GeneralizedLvalue::postmutate( AddressExpr * addrExpr ) {
			return applyTransformation( addrExpr, addrExpr->get_arg(), []( Expression * arg ) { return new AddressExpr( arg ); } );
		}

		Expression * CollapseAddrDeref::postmutate( AddressExpr * addrExpr ) {
			Expression * arg = addrExpr->get_arg();
			if ( isIntrinsicReference( arg ) ) {
				std::string fname = InitTweak::getFunctionName( arg );
				if ( fname == "*?" ) {
					Expression *& arg0 = InitTweak::getCallArg( arg, 0 );
					Expression * ret = arg0;
					ret->set_env( addrExpr->get_env() );
					arg0 = nullptr;
					addrExpr->set_env( nullptr );
					delete addrExpr;
					return ret;
				}
			} else if ( CastExpr * castExpr = dynamic_cast< CastExpr * > ( arg ) ) {
				// need to move cast to pointer type out a level since address of pointer
				// is not valid C code (can be introduced in prior passes, e.g., InstantiateGeneric)
				if ( InitTweak::getPointerBase( castExpr->result ) ) {
					addrExpr->arg = castExpr->arg;
					castExpr->arg = addrExpr;
					castExpr->result = new PointerType( Type::Qualifiers(), castExpr->result );
					return castExpr;
				}
			}
			return addrExpr;
		}

		Expression * CollapseAddrDeref::postmutate( ApplicationExpr * appExpr ) {
			if ( isIntrinsicReference( appExpr ) ) {
				std::string fname = InitTweak::getFunctionName( appExpr );
				if ( fname == "*?" ) {
					Expression * arg = InitTweak::getCallArg( appExpr, 0 );
					// xxx - this isn't right, because it can remove casts that should be there...
					// while ( CastExpr * castExpr = dynamic_cast< CastExpr * >( arg ) ) {
					// 	arg = castExpr->get_arg();
					// }
					if ( AddressExpr * addrExpr = dynamic_cast< AddressExpr * >( arg ) ) {
						Expression * ret = addrExpr->get_arg();
						ret->set_env( appExpr->get_env() );
						addrExpr->set_arg( nullptr );
						appExpr->set_env( nullptr );
						delete appExpr;
						return ret;
					}
				}
			}
			return appExpr;
		}
	} // namespace
} // namespace GenPoly

// Local Variables: //
// tab-width: 4 //
// mode: c++ //
// compile-command: "make install" //
// End: //
